# -*- coding: utf-8 -*-
"""

The basic concept is to read all valid YAML and make it
a versatile config object based on additional parsing
rules intruduced by nao.

Nao should be able to load every YAML file and make it
available for investigation. Only warnings should be issued
if the structure seems odd.

Multiple sources could be projected on each other using a recursive
dictionary structure as it is done in the normal multivariate inner
structure of nao. For different options see anyconfig parsing.

Every state of the nao object tree should return objects that are
both primitive and versatile. There should be local and global
options for plain retreival of objects. The strange setup of
creating objects that both inherit from nao and the respective
base types (or types marked up in the YAML file - dynamic class
creation) is necessary to be able to provide the same (familiar)
structure of source files upon saving!

Attribute like access is allowed to support easy use of
configuration objects and to keep the dict lookup syntax
for layer activation.

Default filters and default search locations are provided
to be able to use Nao as:

from nao import config
OR
import nao
config = nao.load()
"""
# Copyright (C) 2015 Szabolcs Blaga, see LICENSE
# use Python init script properly

# __all__ = ['Nao', 'SimpleFilter', 'load', 'update']

__version__ = '0.1-dev'

import _yaml

odict = _yaml.odict





def no_transform(key):
    return key

def simple_transform(key):
    return '_'+key

class Filter(object):
    pass

class SimpleFilter(Filter):
    """
    translate between nao keys and yaml map keys e.g. 
    macro  ->  _macro
    en  ->  en_US  

    should be able to build filters dynamically based on the
    actual yaml the ano object loads
    """
    def __init__(self, *keys):
        self._keys = keys

    def __contains__(self, key):
        """Implements key in filter functionality"""
        return key in self._keys

    def key(self, index):
        """should return key for a given index"""
        return self._keys[index-1]

    def index(self, key):
        """should return index for a given key (non zero!)"""
        return self._keys.index(key) + 1
    

class BabelFilter(Filter):
    pass

class UnderscoreFilter(Filter):
    pass



default_filters = [
    SimpleFilter('en', 'hu'),
    ]

def load(source, filters=default_filters):
    """
    TODO
    """
    x = _yaml.load(source)
    log.debug(x)

    return Nao(x)


def update(base, other, recursive=True):
    """
    TODO override addition for this purpose!
    """
    if dictlike(base):
        if recursive:
            recursive_update(base, other)
        else:
            base.update(other)
    else:
        raise AttributeError('Update dictionary like objects only.')

    
def recursive_update(base, other):
    """
    Recursively upadtes base with other. Both should be
    dictionary-like mapping types. 

    Returns None as modifies base in place

    This is a module level function and can be used for
    any dictionary like objects including dict, odict
    and naodict

    TODO: check __instancecheck__ for naodict metaclass
    """
    for key, value in other.iteritems():
        if key in base and dictlike(base[key], value):
            recursive_update(base[key], value)
        else:
            base[key] = value
        

def dictlike(*candidates):
    """checking"""
    for c in candidates:
        if not isinstance(c, (dict, naodict)): return False
    return True


def listlike(*candidates):
    """checking"""
    for c in candidates:
        if not isinstance(c, (list,)): return False
    return True


_registry = {} # for special classes that inherit both Nao and the YAML type



def Nao(object):
    """
    It provides an attribute-like access for iteritems
    and a dict like syntay to activate filters.

    naodict
        items accessed by .item_name
            attribute names shield normal dict methods
            normal methods aliased as _name
        filters activated by [*filter_names]
            this shields normal dict access
        for x in naodict iterates through (k,v) pairs
            it uses dict.iteritems

    naolist
        items accessed by [number]
        filters activated by [*filter_names]
        for x in naolist iterates through values

    naostr
        make it python3 only?

    cascading filter?



    """
    def __init__(self, data,
            filters=default_filters,
            mapping_type=naodict):
        self._mapping_type = mapping_type
        self._filters = filters # should be iterable
        self._data = self._parse(data)

    def __getitem__(self, key):
        """
        Should return self with the proper filters 
        activated. Copying would be very perfomance
        intensive.

        >>> config['en'] 
        # returns the particular tree

        """
        # check key as tuple or slice
        for index, filter in enumerate(self._filters):
            if key in filter:
                self._current = list(self._current)[index] = key


    def update(self, other, recursive=True):
        """
        Updates the whole multilayer structure layer by layer
        """
        dat = self._parse(other)
        recursive_update(self._data, dat, recursive=recursive)

    def load(self, stream):
        self._data = self._parse(_yaml.load(stream))

    def _parse(self, data):
        """
        This is the core of the whole fuss
        """
        base = {} # empty on load

        # walk through od recursively and separate it by filters
        # only listlike or dictlike walked others are considered leaves
        index = tuple([0] * len(self._filters))
        base[index] = self._recursive(base, data, data, index)

        return base


    def _recursive(self, base, initial, current, index):
        """ """
        if dictlike(current):
            dres = self._mapping_type()
            keys = self._getkeys(index)
            for k, v in current.iteritems():
                # decide if it is part of the filter structure
                if self._haskey(k):
                    # check is the given key current search index
                    key = k # TODO transform
                    if key in keys:
                        # let the recursion flow in and SKIP! current level
                        return self._recursive(base, initial, v, index)
                        # SHOULD ISSUE warning if a key passed more than once
                    # new key
                    else:
                        new_index = self._addkey(index, key)
                        # if key is contradictory to the current
                        if new_index is None:
                            log.warning('Contradictory key: %s, after %s' % (key, keys))

                        # new_key is already in self._data
                        elif new_index in self._data:
                            log.warning('Multiple data for: %s' % keys)

                        # else start building form the beginning with new_key
                        else:
                            base[new_index] = self._recursive(base, initial, initial, new_index)

                # continue building the tree
                else:
                    dres[k] = self._recursive(base, initial, v, index)
            return dres

        elif listlike(current):
            return [self._recurse(base, initial, item, index) for item in current]

        else:
            return current


    def _getkeys(self, index):
        return tuple([self._filters[f].key(i) for f, i in enumerate(index)])

    def _haskey(self, key):
        for filter in self._filters:
            if key in filter: return True
        return False

    def _addkey(self, index, key):
        position = None
        for i, filter in enumerate(self._filters):
            if key in filter:
                position = i
                break

        if position is None:
            raise KeyError("Cannot add key : %s" % key)

        # only the default could be modified -> contradictory path
        if index[position] != 0:
            return None

        value = self._filters[position].index(key)
        new_index = list(index)
        new_index[position] = value

        return tuple(new_index)

    def save(self):
        """
        To be implemented
        """
        pass

    def __repr__(self):
        """YAML-like reprezentation """


        


'''
# class storage
_class_registry = {}

def get_class(cls):

    # creating nao capable dynamic classes
    if cls not in _class_registry:
        _class_registry[cls] = type('Nao'+cls.__name__, (Nao, cls), {})

    return _class_registry[cls]


class Nao(object):
    """
    This represents a versatile object

    Multilingual support through BabelFilter

    Representation has both YAML and various database backends

    Nao creates different types on the fly. These classes has some
    extra properties starting with `_nao_`. So ideally as a main rule
    classes with attributes starting with `_nao_` could mess up nao other
    classes are OK.

    Every node could have a value (its original class) and many
    subnodes based on the Nao object. Subnodes exposed through attributes
    unless they hide functionality in that case a warning is issued. This
    makes it more readable e.g. in template usage.

    Active filters could be set through the __getitem__ functionality, so
    nao objects are hiding original __getitem__ functionality by default
    for string keys as the filter names should always be strings.
    """

    def __init__(self, filters=default_filters, allow_attributes=True):
        """
        allow_attributes
            If True string dictionary keys restricted to the non-unicode naming
            conventions of Python variables are exposed as attributes. This has
            a beutifying effect of using Nao objects in certain contexts, e.g.
            templating languages.

        Layers of data organized by filters are stored in _nao_dict with
        a key constructed as a tuple of the index of the keys in each
        filter.

        Filters should have unique keys!
        """
        self._nao_config = {}
        # self._nao_filters = FilterContainer(filters) # share by reference
        self._nao_odict = odict() # copy on filter
        super(Nao, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        """
        Hiding functionality by default for keys that are exist in any filters
        for the object.
        """
        if (
            self._nao_config.getdefault('filters_prevail', True) and
            self._nao_filters.has_key(key)
            ):
            # should replace itself if immutable so need to return a
            # reference to the value (or a copy)
            return self._nao_filter(key) # propagate false?
        else:
            super(Nao, self).__getitem__(self, key)

    def __getattr__(self, key):
        """
        Based on nao settings expose functionality if it is not hiding
        something else. Otherwise nao functionality is only accessible via
        the `_nao_*` names.
        """
        if (
            self._nao_config.getdefault('allow_attributes', True) and
            key in self._nao_odict[self._nao_active]
            ):
            return self._nao_odict[self._nao_active][key]
        else:
            super(Nao, self).__getattr__(self, key)

    def __setattr__(self, key, value):
        """
        Dangerous for two reasons
            could render nao useless
            autosave is very resource intensive
        Need to replace mutable objects
        """
        super(Nao, self).__setattr__(key, value)



    def _nao_filter(self, key, propagate=True):
        """
        Usually called as filter('name') and it propagates through
        the whole tree and sets every object to the corresponding
        value. This means that the immutable nodes should be recreated
        with the Nao object copied over.
        """
        pass

    def _nao_load(self, input):
        """
        read input as YAML and update self
        """
        self.update()

        cls = self.__class__






def dump():
    # dynmically dump if the data changes
    # thread-safety?


def default_language_test_callback(string):
    """decide weather a given string is considered a locale"""


def default_source_list():
    """
    The list of sources that nao implementation tries by default
    to load Nao objects

    `nao.yaml` at the current directory and that file could have
    more information regarding autoload behaviour through
    `_nao` configuration options

    all yaml files stored in the `config` subdirectory
    """


def load(source=default_source_list):
    """
    Create an empty Nao object and call load on import
    """
    return Nao().load(source=source)


class FilterDescriptor(object):
    """
    """
    def __init__(self, names, transform):
        self._names = names
        self._transform = transform

    @property
    def names(self):
        """return those name identifier that are part of this filter"""
        return self._names

    def transform(self, name):
        """
        Mapping betwwen exposed filter names and effective names

        E.g.    names: startswith('_')
            transform: name[1:]
        """
        return self._transform(name)



class BabelFilter(FilterDescriptor):
    """Need to have connection with the Babel package or directly with the CLDR"""
    pass

class SimpleFilter(FilterDescriptor):
    pass


default_filters = {
    #'lang': BabelFilter(),
    # 'flag': SimpleFilter(name_prefix='_')
}





class RecursiveDictionary(dict):
    """RecursiveDictionary provides the methods rec_update and iter_rec_update
    that can be used to update member dictionaries rather than overwriting
    them.

    Created 2009-05-20 by Jannis Andrija Schnitzer (jannis@itisme.org)
    https://gist.github.com/114831
    Modified 2012-12-01 by Szabolcs Blaga (szabolcs@syrus.hu)
    """
    def recurse(self, other, **third):
        """Recursively update the dictionary with the contents of other and
        third like dict.update() does - but don't overwrite sub-dictionaries.

        Example:
        >>> d = RecursiveDictionary({'foo': {'bar': 42}})
        >>> d.rec_update({'foo': {'baz': 36}})
        >>> d
        {'foo': {'baz': 36, 'bar': 42}}
        """
        try:
            iterator = other.iteritems()
        except AttributeError:
            iterator = other
        self.iter_recurse(iterator)
        self.iter_recurse(third.iteritems())

    def iter_recurse(self, iterator):
        for (key, value) in iterator:
            if key in self and isinstance(self[key], dict) and isinstance(value, dict):
                self[key] = RecursiveDictionary(self[key])
                self[key].recurse(value)
                self[key] = dict(self[key]) # avoid using multi-level recursive dictionaries
            else:
                self[key] = value

    def __repr__(self):
        return super(self.__class__, self).__repr__()

"""
config = RecursiveDictionary()
for f in [path(n+'.yaml') for n in ('global', 'local')]:
    if os.path.isfile(f):
        c = yaml.safe_load(codecs.open(f, 'r'))
        if c is not None: config.recurse(c)

config = dict(config)
"""

# this is the default config object and the modul level
# functions are working on this e.g. load, etc.
# nao = Nao(autoload=default_source_list)
# setting modul default based on nao.yaml/nao.yml
# from nao import nao




'''






import logging
import datetime

from sqlalchemy import *

class DBHandler(logging.Handler):
    """
    allows the logging into an SqlAlchemy ORM
    
    Uses an engine and sets up separate threadsafe
    connection to that engine
    """

    def __init__(self, engine):
        """

        """
        super().__init__()

        # create table if needed

        # initialize session
        self.session = None

    def formatDBTime(self, record):

        # WARNING modifies created
        record.created = datetime.datetime.fromtimestamp(record.created)


    def emit(self, record):

        # Use default formatting
        self.format(record)

        exc_formatter = logging._defaultFormatter.formatException

        record.exc_text = exc_formatter(record.exc_info) if record.exc_info else ''

        self.session.add(LogRecord(**record.__dict__))
        self.session.commit()

    


class LogRecord(Base):

    inserted =      Column(String, default=datetime.datetime.now)
    created =       Column(Datetime)
    name =          Column(String)
    level =         Column(String)
    module =        Column(String)
    func_name =     Column(String)
    line_no =       Column(String)
    thread =        Column(String)
    thread_name =   Column(String)
    process =       Column(String)
    message =       Column(Text)
    args =          Column(Text)
    exception =     Column(Text)
    

    def parse_message():
        """
        some logic to parse message into YAML and then
        annotate it in the viewer...
        """





import textwrap

class fake_host(object):
    """
    use as:

    with fake_host('syrus.hu'):
        server.serve(app)
    """

    hosts_files = {
        'posix': r'/etc/hosts',
        'nt': r'C:\Windows\System32\drivers\etc\hosts',
    }

    def __init__(self, host, ip='127.0.0.1'):
        self.insert = textwrap.dedent("""
        
        # Automagically inserted line for OAuth testing follows:
        {} {}

        """.format(ip, host))
        app.logger.debug(self.insert)

        # split between Windows and Linux (need sudo for that)
        try:
            self.hosts_file = hosts_files.get(os.name, None)
        except:
            raise ValueError('OS not supported')

    def __enter__(self):
        with open(hosts, 'r+') as h:
            # save the current in the tool and restore it on server stop
            x = h.read()
            print(x)
            if x.find(insert) == -1:
                h.truncate(0)
                h.seek(0)
                h.write(x + insert)
                print(insert)

    def __exit__(self, exc_type, exc_value, traceback):
        # restore hosts (this is not runninbg as a standalone script in production environment!)
        with open(hosts, 'r+') as h:
            # search for line with special comment and replace
            x = h.read()
            print(x)
            y = x.replace(insert, '')
            print(y)
            h.truncate(0)
            h.seek(0)
            h.write(y)


# -*- coding: utf-8 -*-
"""
this should be a basic package that
reads everything from a directory
structure (data lazy) and provide
a flexible object model to access
all the resources

- data files (rst, yaml, csv, xlsx, etc)
    rst needs support for :ref: inside nao datastructure
    add rst parsing to yaml files, also support cross yaml referencing with rst style tags
- database connections (sqlite, special yaml files using, sqlalchemy models?)
- automatically serve those files as a webserver
- python files for code support (no relative import only through nao.self?)
    - decorators to provide functionality outside (e.g. auth, etc)
- template files made available automatically (in nao.current.template)

later multilayer support should be added
the main resaon for that is supporting
multilingualism (also production, testing separation)

multilingualism is supported by babel negotiatble tags(e.g. en_US)

mode separation is set up based on _identifier style tags

"""
from __future__ import unicode_literals # no need in python3 stack

# basic logging
import os
import sys
import logging
import time
import functools
import codecs
import re
import collections

# from future.utils import python_2_unicode_compatible

from ._yaml import load as yaml_load, odict, naodict

# no handler is added
rootlogger = logging.getLogger()  

def logger(name,
        level='debug',
        dict_config=None,
    ):
    """
    For logging in modules outside of nao use::
    log = nao.logger(__name__)
    """
    # process dict config

    # not necessary since Python2.7 accept string levels
    levels = {
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }
    lvl = levels[level.lower()]

    log = logging.getLogger(name)
    log.setLevel(lvl)

    log.root = rootlogger

    return log

# default logger for nao
log = logger(__name__) 


op = os.path


def load(source=None):
    """
    main entry point, loads everything from a given point
    of the file system (usually __file__)

    builds a tree based on file names and parse
    files based on extension and alphabetical
    order

    .yaml
        regular python data
        with some type of nao reference type available
        maybe all text fields interpreted as rst is a good idea
        add protocol handling as well (file://, ext://, oracle://, ...)
        see `logging.config.dictConfig` for reference

    .rst
        parsed as in sphinx docs into raw html
        :nao:`something` should be available

    .csv, .sql, .xls, .xlsx
        deferred datatable
        sql should run on the default db connection
        defined by nao-tree (with special .nao files)
        if nothing defined .nao-s own default sqlite
        db should be used (cashing, etc also in that)
    
    .nao
        special resource described in YAML
        e.g. database connaction, RPC service, aPI calls, etc.
        return a Python object???
        similar .rst's and .yaml's ext directive

    .html, .jinja
        parsed as jinja templates by default
        the object returned in load is available by the
        name nao in tempate global (think it over more)

    .py
        parsed as python code and an object tree
        is build (I know this is going to be controversial...)

    .jpg, .png, .mp3, .wav, .avi, ...
        parsed as media resource (embed in template)


    read system information as well and do
    everything in a multilayer and cascading
    manner

    returns a root element
    """
    if source is None:
        # doubles the check for existence, but avoids strange errors
        if op.exists('nao'):
            source = 'nao'

        # TODO check for nao.yaml, nao/, config.yaml, config/
        # in some order defined as a cutomizable parameter

        # finally load current directory as a config directory

    if op.exists(source):
        if op.isdir(source):
            log.debug("Loading from directory recursively")
            pass
            # TODO implement recursive directory loading
            # see Naopath.filter
        else:
            log.debug("Loading from single file")
            with codecs.open(source, encoding='utf8') as f:
                # TODO add content negotiation here
                return yaml_load(f)

    else:
        log.debug("Loading from string")

        # TODO create yaml naodictloader

        return yaml_load(source)
    # if source == None, then try to load [cwd]/nao

    # get directory based on source

    # create root_node of current directory
    # walk recursively, skip everything that starts with _
    # create nao* object-tree

    # nao.path(data.config.general)
    # str(data.config) # returns a path if it is a file/directory
    # nao.url(naobject :)


def separate(data, values):
    """
    recursively traverse data tree (mappings, lists)
    and yield that many different data stucture as the
    length of values
    """

    for val in values:

        yield val, _filter(data, val)


# from copy import deepcopy

def _filter(data, val=None):

    if isinstance(data, dict):
        if val is not None and val in data:
            return _filter(data[val])

        else:
            r = {}
            for k in data:
                r[k] = _filter(data[k], val)
            return r

    elif isinstance(data, list):

        return [_filter(i, val) for i in data]

    else:
        return data


# from .utils import *
def ordered(*args, **kwds):
    """
    Constructs an `odict` from `args` and `kwds`
    From PYthon 3.6 dict preserve order
    """
    od = odict()
    for a in args:
        od[a] = kwds[a] if a in kwds else None
    return od


u1reg = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
u2reg1 = re.compile('(.)([A-Z][a-z]+)')
u2reg2 = re.compile('(a-z0-9])([A-Z])')

def underscore(text, method='single_reg'):
    """
    Largely based on Stack Overflow.
    Two methods are only implemented out of curiosity.

    Converts `CamelCase` or `camelCase` to under_score style.
    Able to handle `camelAndHTTPResponse` as `camel_and_http_response`.
    Avoids multiple underscores, so `under_Score` remains `under_score`
    does not become `under__score`.
    """

    if method == 'single_reg':
        return u1reg.sub(r'_\1', text).lower().replace('__', '_')

    if method == 'double_reg':
        temp = u2reg1.sub(r'\1_\2', text)
        return u2reg2.sub(r'\1_\2', temp).lower().replace('__', '_')

    raise ValueError("Method not recognized: {}".format(method))


def flatten(d, parent_key='', sep='.'):
    """
    FROM http://stackoverflow.com/questions/6027558/flatten-nested-python-dictionaries-compressing-keys
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def unflatten(d):
    """
    TODO implement
    """


# from .operators import *
def back():
    """
    operator on nao nodes

    it could be used in a special interpreter later
    """

def path():
    """operator on nao nodes"""


# TIMING CONTEXTMANAGER AND DECORATOR
def benchmark(arg):

    # normal use returns a context manager
    if not callable(arg):
        return Benchmark(str(arg))

    # decorator use: wrapped into a benchmark context
    func = arg
    @functools.wraps(func)
    def wrapper(*args, **kwds):
        with Benchmark('FUNC ' + func.__name__):
            return func(*args, **kwds)

    return wrapper

class Benchmark(object):
    """
    Based on something similar found long time ago on the internet
    """
    def __init__(self, name):
        self._name = name
        self._time = None

    def __enter__(self):
        self._begin = time.time()
        log.info("Benchmark <{}> started ...".format(self._name))

    def __exit__(self, exc_type, exc_value, traceback):
        self._time = time.time()-self._begin
        log.info("Benchmark <{}> finished in: {}".format(self._name, self.time))
        return False

    @property
    def time(self, raw=False):
        if time is None:
            raise ValueError('Not run yet')

        return self._time if raw else self.format_time(self._time)
    

    @staticmethod
    def format_time(seconds):
        """
        TODO use `babel.dates.format_timedelta` with
        `datetime.timedelta(seconds=...)`

        """
        t = []
        
        if seconds > 60*60:
            hours, seconds = divmod(seconds, 60*60)
            t.append("{:.0f}h".format(hours))
        
        if seconds > 60:
            mins, seconds = divmod(seconds, 60)
            t.append("{:.0f}m".format(mins))

        t.append("{:.3f}s".format(seconds))

        return ' '.join(t)




def _os_join(*bits):
    log.debug('join {}'.format(bits))
    return op.normpath(op.join(*bits))

def _url_join(*bits):
    return '/'.join(*bits)


def path(*base):
    """shortcut to prevent direct class usage"""
    return NaoPath(*base)

# @python_2_unicode_compatible
class NaoPath(object):
    """
    use a naodict to navigate through directory structure
    and __call__ to return a path string

    base = MagicPath('base_path')
    base.config == base['config'] # another MagicPath
    base.__up__ ?? base['..']
    base.config()  # base_path/config/
    base.config('my.conf')  # base_path/config/my.conf

    normpath is for reundant separators and up-level references
    realpath is to eliminate symbolic links
    """
    def __init__(self, *base):
        b = op.realpath(op.normpath(op.join(*base)))
        # TODO fails for non-existent dir!!!
        self.base = op.dirname(b) if op.isfile(b) else b


    def __getitem__(self, key):
        # log.debug(key)
        # log.debug(type(key))
        if isinstance(key, tuple):
            return self.__class__(self.__call__(*key))
        else:
            return self.__class__(self.__call__(key))


    def __call__(self, *bits):
        # .normpath
        p = op.normpath(op.join(self.base, *bits))
        if not op.exists(p):
            log.warning("Non-existent path: {}".format(p))
        return p


    def __str__(self):
        return self.base


    def filter(self, expression=None, filetype=None, recursive=True, **kwds):
        """
        get certain files from a path
        """
        p = self()

        if not op.exists(p) or not op.isdir(p):
            return

        for root, dirs, files in os.walk(p, **kwds):

            if filetype:
                for f in files:
                    if f.endswith('.'+filetype):
                        yield op.join(root, f), f[:-len(filetype)-1]

            if not recursive:
                break


    # def __enter__():
    #     sys.path.insert(0, self.base):

    # def __exit__():
    #     sys.path.remove(self.base):


_nao_path = NaoPath(__file__)  # path for the nao package


def read_dir_raw_text(location, filter=None, encoding='utf8', recursive=False):

    for root, dirs, files in os.walk(location):
        for name in files:
            if filter is not None:
                if not filter(name): continue
                
            with open(op.join(root, name), encoding=encoding) as f:
                data = f.read().strip()  # strip might not adequate here
                yield name, data

        if not recursive: break # no recursion into subdirs


# read_sql from load oracle should come here somewhere


# nao.ext.sphinx

def make_html(doc_path, **kwds):

    return SphinxDoc(doc_path, **kwds)

    
    

class SphinxDoc(object):
    """Wrapping a sphinx documentation"""
    def __init__(self, doc_path, **kwds):
        """ """
        self._conf = {
            'master_doc':'index',
            'html_theme':'alabaster',
        }
        self._conf.update(kwds)
        self._doc_path = doc_path

        # TODO pop special args like build_path
        self._build_path = path(doc_path)('_build', 'html')


    def build(self):
        """ """
        #preparse options:
        conf = self._conf.copy()
        for k,v in conf.items():
            # list is comma separated
            if isinstance(v, list):
                conf[k] = ','.join(v)
            # dict is split
            if isinstance(v, dict):
                for k2,v2 in v.items():
                    conf[k+'.'+k2] = v2
                del conf[k]
            # boolean is 0 or 1
            if isinstance(v, bool):
                conf[k] = '1' if v else '0'


        self._opts = ' '.join(['-D '+k+"='"+v+"'" for k,v in conf.items()])
        # conf = "master_doc='index' -D html_theme='scrolls'"
        self._cmd = "sphinx-build -b html -C {} {} {}".format(self._opts, self._doc_path, self._build_path)

        os.system(self._cmd) # include output in debugging

        return self

    @property
    def command(self):
        return self._cmd

    @property
    def build_path(self):
        return self._build_path

    @property
    def doc_path(self):
        return self._doc_path

    @property
    def options(self):
        return self._opts

    @property
    def config(self):
        return self._conf
    


#################
# MOVE THIS TO nao.ext.web or something :D

class App(object):
    """
    This creates a WSGi application from a file
    structure based on the input parameters and
    app-level config (the config also read based
    on input parameters)

    Sensible defaults are:

    config/
        config directory
        files read: .yml, .yaml, .json, .cfg, .ini (=> incorporate anyconfig)
        namespace:
            directories
            # file names  (len < 24 and alpha start and single dot)
            # // ?? //
            # files only for concept separation

    handler/
        handler directory
        default style: cherrypy
        files parsed: .py
        use __root__ instead of default

    model/
        model directory
        default style: sqlalchemy
        files parsed: .py


    Alternative named setup structures:

    single-hierarchy
        parsing is in a single directory structure
        the file type is used to ditinguish separate
        elements (config, template, handler, model)

        all directories above are substituted for
        the root directory (e.g.: ``./``)

        Needs: import __init__ as root

    sphinx-kb
        A knowledge base type static page using sphinx

        sphinx-build -C bkd/static bkd/static/.sphinx
            http://sphinx-doc.org/invocation.html
        
        if no toctree, then ...
            use a generated __root__.rst to include every
            document with the glob option:
            http://sphinx-doc.org/markup/toctree.html
        
        webbrowser.open( url, new=0 ) 

    sphinx-blog
        A blog engine using sphinx :)


    WARNING make this a WSGI application
    """



    def __init__(self, root,
            setup='nao',
            templating='jinja2',
            modelling='SQLAlchemy',
            ):
        # derive directories from root with nao.path()
        root = os.path.split(os.path.realpath(root))[0]
        self.path = MagicPath(root)  #: path for the application

        self.setup = setup
        self.handlers = object()

        app_root = self.path()
        print(app_root)
        # TODO
        for root, dirs, files in os.walk(app_root):
            for d in dirs:
                if d.startswith('.'):
                    dirs.remove(d)

            for f in files:
                if not f.startswith('.') and not f.startswith('_'):
                    name,ext = os.path.splitext(f)
                    ext = ext[1:]
                    self._process(root, name, ext)
                    

    def _process(self, root, name, ext):
        """
        # applyprocessors
            # handler
                # spam = __import__('spam', globals(), locals(), [], 0)
            # config
            # model
            # template

            
        islink, join, isdir = path.islink, path.join, path.isdir

        # We may not have read permission for top, in which case we can't
        # get a list of the files the directory contains.  os.path.walk
        # always suppressed the exception then, rather than blow up for a
        # minor reason when (say) a thousand readable directories are still
        # left to visit.  That logic is copied here.
        try:
            # Note that listdir and error are globals in this module due
            # to earlier import-*.
            names = listdir(top)
        except error, err:
            if onerror is not None:
                onerror(err)
            return

        dirs, nondirs = [], []
        for name in names:
            if isdir(join(top, name)):
                dirs.append(name)
            else:
                nondirs.append(name)

        if topdown:
            yield top, dirs, nondirs
        for name in dirs:
            new_path = join(top, name)
            if followlinks or not islink(new_path):
                for x in walk(new_path, topdown, onerror, followlinks):
                    yield x
        if not topdown:
            yield top, dirs, nondirs

        """
        if ext == 'py':
            print(root+'-:-'+name+'-:-'+ext)
            # WARNING not threadsafe!
            sys.path.insert(0, root)
            sys.path.pop(0)


            

    def run(self, name, open=True):
        """


        """
        if name != '__main__': return

        if self.setup == 'sphinx-kb':
            # run sphinx
            p = self.path()

            # ' '.join(*['-D '+k+"='"+v+"'" for k,v in conf.iteritems()])
            conf = "master_doc='index' -D html_theme='scrolls'"
            os.system("sphinx-build -C -D {} {} {}/.sphinx".format(conf, p, p))
            
            self.base_url = self.path('.sphinx', 'index.html')
            
            self.open()


            # TODO setup cherrypy to serve static files from .sphinx/_build/html 
            


        else:
            self.base_url = 'http://localhost:8080'

            import cherrypy

            # cherrypy.config.update({'environment': 'embedded'})

            class HelloWorld:
                def index(self):
                    return "Hello world!"
                index.exposed = True

            if open:
                cherrypy.engine.subscribe('start', self.open)

            cherrypy
            cherrypy.quickstart(HelloWorld())


    def open(self):
        print('called')
        import webbrowser

        webbrowser.open(self.base_url, new=0)



class Server(object):

    def __init__(self):
        pass

    def serve(self, app):
        """
        The app could have configuration for the server, e.g.:

        port, ip, mount_point, etc

        The server could try to serve many apps in the same
        Python interpreter
        """
        pass

        # implement a cherrypy server start here :)

        # webbrowser.open( url, new=0 )  # open localhost



# read config before
server = Server()


auth = object()  # auth tool with OAuth support through requests_oauthlib
# add auth.all, auth.user and auth.groups() shortcut auth()

request = object()  # the dynamically changing request object
# expose cherrypy request object

db = object()  # a threadsafe SQLAlchemy session (as deault on SQLite), version aware 

expose = object() # a default decorator to add handlers



import numpy as np
import decimal as dec

import itertools

from inspect import isclass, getmro

from . import logger, benchmark, odict

from sqlalchemy.types import DateTime, Date
from sqlalchemy.types import Integer, BigInteger, Float, Numeric, Boolean, Enum
from sqlalchemy.types import String, Text, Unicode, UnicodeText
from sqlalchemy.schema import Column
from sqlalchemy.orm import sessionmaker
from sqlalchemy.inspection import inspect
from sqlalchemy.ext.declarative import declared_attr

log = logger(__name__)


# TODO need to centralize resource loading (query and Excel)
# to ensure data integrity


class LoaderError(Exception): pass

class LoadableMixin(object):
    """
    This is the Basic mixin that performs loading functionality

    This should be created similarly to declarative base otherwise
    shared among all model classes in the same interpreter process

    """
    __loader__ = None  #: override class attribute in subclasses
    __loader_args__ = {} #: loading arguments shared by subclasses

    @classmethod
    def load(cls, params=None, ranges=None):
        """
        """
        metadata = get_metadata(cls)

        for pars in cls._parse_ranges(ranges, params):

            s = cls._get_session(pars, metadata)

            res = cls._load(pars)

            s.bulk_insert_mappings(c, res)

            s.commit()
            s.close()


    @classmethod
    @benchmark
    def load_all(cls, params=None, ranges=None):
        """

        `params`
        These params are used to prepare sql-statements 


        `ranges`
        Used to iterate through ranges (preferably an `odict`)
        """
        # TODO somehow stop relying on table names

        loader_classes, metadata = sorted_subclasses(cls)
        log.debug("loader classes: {}".format(loader_classes))
        log.debug("metadata: {}".format(metadata))

        for pars in cls._parse_ranges(ranges, params):

            s = cls._get_session(pars, metadata)

            cache = cls._load_all(loader_classes, pars)

            # bulk insert chould operate separately by the cache
            for key, res in cache.items():
                c = loader_classes[key]
                s.bulk_insert_mappings(c, res)

            s.commit()
        
            # TODO implement row-by-row insert for debugging
            # for row in res:
            #     s.bulk_insert_mappings(self.cls, [row])
            #     s.commit()
            s.close()


    @classmethod
    def _load(cls, params=None):
        """
        Get classes, no ranges, no session
        
        return cls objects loaded from engine

        """
        # TODO use only __loader__args__ instead
        if isinstance(cls.__loader__, Loader):
            # delegate task to loader object
            log.debug('Loading for model class: {}'.format(cls))

            return cls.__loader__.load(cls, params)

        else:
            # rasie some more sensible error
            log.info("{} is not loadable!".format(cls))
        


    @classmethod
    def _load_all(cls, loader_classes, params):
        """
        Get classes, no ranges, no session

        Implement implicit foreign keys
        """
        # TODO use cache only if there is implicit foreign
        cache = odict()
        implicit_foreign = []

        for tname, klass in loader_classes.items():

            # relies on _load here
            log.debug("Loading classes for {} and {}".format(tname, klass))
            res = klass._load(params)

            cache[tname] = res

            for r in inspect(klass).relationships:

                # TODO what if related class is not in loader classes
                rtbl = r.target.name

                # check if target allows imlicit foreign key addition
                if (
                    rtbl in loader_classes and
                    issubclass(loader_classes[rtbl], ImplicitForeignMixin)
                    ):
                    # target should already exist in the cache because of sorted_tables
                    implicit_foreign.append((klass, r))
                    

            # TODO For DiffLoader check existence here
            # New (add new items with insert_date)
            # Match (implicit records should drop, others update_date)
            # Delete (no actual deletion only delete_date)


        # making implicit foregin key happen
        # TODO many to many??? association proxies???
        # TODO not null howto?

        log.debug("Searching implicit foreign keys...")

        def get_index(table, keys):
            """
            TODO might implement caching of indices here
            """
            index = []
            for row in cache[table.name]:
                index.append(tuple([row[k] for k in keys]))
            return index

        for klass, relation in implicit_foreign:
            # source is the table with foreign_keys to target

            fks, pks = tuple(zip(*[(l.name, r.name) for l, r in relation.local_remote_pairs]))
            # print(fks, pks)
            # target is accepting new records with implicit_foreign flag
            target_index = get_index(relation.target, pks)
            # print(target_index)
            target_class = loader_classes[relation.target.name]

            for row in cache[klass.__tablename__]:
                ind = tuple([row[k] for k in fks])
                if any([i is None for i in ind]): continue

                # print(ind)
                if ind not in target_index:

                    log.warning("Implicit Foreign {} ({}) from table {}".format(
                        relation.target.name,
                        ", ".join([str(i) for i in ind]),
                        klass.__tablename__
                    ))

                    new_row = dict((k,v) for k,v in zip(pks, ind))
                    new_row[IMPLICIT_FOREIGN] = True
                    cache[relation.target.name].append(new_row)

                    # update target_index
                    target_index.append(ind)
                    # target_class.add_implicit_foreign(ind) # return a dict...

            # read all data from klass and check if already in target_data

        return cache


    @classmethod
    def _parse_ranges(cls, ranges, params):

        if ranges is not None:

            names, iterables = ranges.keys(), ranges.values()

            # initialize all parameters once with a Carthesian product
            prod = itertools.product(*iterables)

            for vector in prod:
                
                d = dict(zip(names, vector))

                log.info("LOADING values: {}".format(d))

                p = params.copy()
                p.update(d)


                log.debug("loading params: {}".format(p))
                yield p


        # simply loading with params
        else:
            log.debug("loading params: {}".format(p))
            yield params



    @classmethod
    def _parse_data(cls, data,
        return_raw_dict=True,
        position_based=True,
        only_one_row=False,
        ):
        """
        returns a list of dicts

        validated and coerced to the corresponding model
        """

        # TODO remove columns with special usage
        columns = cls.__table__.columns
        log.debug("Parsing column {}".format(columns))

        res = []

        if only_one_row:
            data = [data]

        # TODO if slow then pandas might use some C magic for this but I doubt (should check)
        i, j = 0, 0  # if the resultproxy has no rows...

        for i,row in enumerate(data):
            # allow extra columns at the end with default values or null
            if position_based and len(row) > len(columns):
                raise ValueError("Degenarate raw (col count mismatch)")

            new_row = {}

            # log.debug("The {} raw row: {}".format(i, row))

            for j,col in enumerate(columns):

                if position_based:
                    # it is OK to have extra columns at the end
                    if j >= len(row): break
                    val = row[j]

                elif col.name in row:
                    # name based could have extra rows anywhere
                    val = row[col.name]
                else:
                    log.warning("Column {} missing for row {}".format(col.name, i))
                    continue

                if val is not None:
                    try:
                        val = cls._parse_value(val, col)
                    except:
                        log.error("row {}, col {}, value [{}], type {}".format(i, j, row[j], type(row[j])))
                        log.error("COLUMN {} ({}) pk: {}".format(col.name, col.type, col.primary_key))
                        raise

                # log.debug("Value parsed for {}: {}".format(col.name, val))
                new_row[col.name] = val

            if return_raw_dict:
                res.append(new_row)
            else:
                res.append(cls(**new_row))

        log.debug("Parsed {} rows with {} columns".format(i+1, j+1))

        if only_one_row:
            return res[0]
        else:
            return res


    @staticmethod
    def _parse_value(val, col):

        # TODO check foreign keys for uniqueness and no None!
        # if c.foreign_keys:
        #     log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))

        # TODO think about primary key column check
        # TODO think about unique constraint check

        # determine numerical input types
        raw_numeric_types = (int, dec.Decimal, float)

        # SQLalchemy types categorized
        numeric_types = (Integer, BigInteger, Numeric, Float)
        date_types = (Date, DateTime)
        string_types = (String, Unicode, Text, UnicodeText)
        bool_types = (Boolean, )

        # very lazy approach to bool casting
        true_vals = (True, 'Y', 'I', 1, '1')  
        # false_vals = (False, 'N', 'N', 0, '0')

        # numeric types only go to numeric types
        if isinstance(val, raw_numeric_types):
            if isinstance(col.type, numeric_types):
                pass
                # TODO check for possible precision loss
                # log.debug("Casting numeric to numeric")
                # val = dec.Decimal(str(val)) OR let the engine do that :)

            elif isinstance(col.type, string_types):
                log.warning("Casting numeric to string")
                orig = val
                val = str(val)
                if col.type.length is not None and len(val) > col.type.length:
                    val = val[:col.type.length]
                    # remove unnecessary end zeros...
                    if orig != type(orig)(val):
                        raise ValueError("COLUMN {} would have been truncated {}".format(col.name, val))


            else:
                raise ValueError("Casting numeric to non-numeric!")

        # str and anything except for numeric and None
        else: 
            if isinstance(col.type, date_types):
                val = pd.to_datetime(val)
                # TODO stp using pd
                # TODO unrealistic date (what is realistic? (today +/- 120 years)
            # series.between(left, right)

            elif isinstance(col.type, numeric_types):
                # Everything casted to Decimal to avoid precision problems, only one conversion occurs at the SQL side
                if val.startswith('0'):
                    log.warning("Converting '0' starting strig to numeric")
                # check dtypes and issue warnings
                val = dec.Decimal(val)

            elif isinstance(col.type, string_types):
                val = val.strip()
                # callback mechanism for maximum string length in column! (self._callback_string_length)
                # check string and unicode columns length attribute
                if col.type.length is not None and len(val) > col.type.length:
                    raise ValueError("COLUMN {} would have been truncated {}".format(col.name, val))


            elif isinstance(col.type, bool_types):
                val = val in true_vals

        return val


    @staticmethod
    def _get_session(params, metadata):

        te = params.pop('target_engine', None) or metadata.bind

        if not te:
            raise ValueError("no target engine")

        log.debug('Target engine: {}'.format(te))
        
        return sessionmaker(bind=te)()


IMPLICIT_FOREIGN = 'implicit_foreign'

class ImplicitForeignMixin(LoadableMixin):
    """
    This class is mainly an indicator to initiate
    impliciet loading mechanism based on the foreign keys
    of other `Loadable` classes of the same `LoadableBase`.

    Adds a `Column` by the name defined in `IMPLICIT_FOREIGN`
    constant.
    """


def anon(cls):
    return None


# TODO remove column from load iteration!
# PATCH using `declared_attr` moves it to the end otherwise
# `column_instance._create_order` defines or need to flag mixin
# columns otherwise


setattr(ImplicitForeignMixin, IMPLICIT_FOREIGN, declared_attr(lambda cls: Column(Boolean, default=False)))


def sorted_subclasses(cls):
    """
    the collecting mechanism is that the classes
    should be inherited from the caller

    TODO change that they should inherit from
    Base and have a __loader__ attribute
    create Base.metadata.load_all() somehow
    """

    # get all subclasses that should have a __tablename__:
    lst = odict()

    for c in subclasses(cls):
        print('scls', c)
        if hasattr(c, '__tablename__'):
            lst[c.__tablename__] = last_cls = c
            

    # get metadata
    metadata = get_metadata(last_cls)
    
    loader_classes = odict()

    if metadata:
        for t in metadata.sorted_tables:
            if t.name in lst:
                loader_classes[t.name] = lst[t.name]

        return loader_classes, metadata

    else:
        raise ValueError("No metadata found :(")


def subclasses(cls):
    for subclass in cls.__subclasses__():
        yield subclass
        yield from subclasses(subclass)


def get_metadata(cls):
    """
    _declared_class_registry might be safer...
    """
    for c in cls.__mro__:
        print('mro', c)
        if hasattr(c, 'metadata'):
            return c.metadata



import pandas as pd  # TODO REMOVE pandas dependency!


class Loader(object):
    """
    Base class for all the loading functionality

    This could be an abstract base clas
    """
    pass


class ParamLoader(Loader):
    
    def __init__(self, **kwds):
        """
        """
        self.kwds = kwds


    def load(self, model_class, params):

        self._get_config(model_class, params)

        self.model_class = model_class
        self.table = model_class.__table__

        # debug usually from load or load_all call
        self.debug = self.cfg.pop('debug', False)

        self.tbl = model_class.__table__
        # log.debug(tbl)

        # call class specific loader mechanism
        return self._load()


    def _get_config(self, cls, params):
        x = cls.__loader_args__.copy()
        x.update(self.kwds)
        x.update(params)
        self.cfg = x




class SQLLoader(ParamLoader):


    @staticmethod
    def prepare_sql(sql, **params):
        sql = sql.strip()
        if sql.endswith(';'): sql = sql[:-1]
        sql = sql.format(**params)
        return sql


    def _load(self):

        cfg = self.cfg
        # get engine
        engine = cfg.get('source_engine')

        # INFO no defaults needed as if there is it should be in the
        # SQL (otherwise a cfg called params could be used)
        sql = cfg.get('load_sql')
        # create prepare sql class which use parameter inclusion
        # might wanna check sqlalchemy syntax for that!
        sql = self.prepare_sql(sql, **cfg)
        # log.debug(sql)
        
        # read sql
        res = self.model_class._parse_data(engine.execute(sql))
        # res = self._read_sql_slow(sql, engine)

        # search duplicates
        if self.debug:

            import pandas as pd

            df = pd.DataFrame(res)
            pks = [c.name for c in self.table.columns if c.primary_key]
            log.debug("primary key: {}".format(pks))
            dpl = df.duplicated(subset=pks, keep=False)
            if dpl.any():
                log.debug("DUPLICATE: {}".format(df.duplicated(subset=pks).any()))
                df[dpl].to_csv('dupidup.csv')
                # TODO make output filename more sane

        # TODO add callback mechanism for other tables
        # for c in self.cls.__table__.columns:
        #     if c.foreign_keys:
        #         log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))

        return res


    


class ExcelLoader(ParamLoader):
    

    def _load(self):
        """
        Excel load mechanism using pandas for quick implementation
        """

        return

        cfg = self.cfg

        path = cfg.get('path')
        sheet = cfg.get('sheet', None)
        cols = cfg.get('columns', None)


        import pandas as pd

        xl = pd.ExcelFile(path)
        # print(xl.sheet_names)

        df = xl.parse(sheet if sheet is not None else 0)

        # filter by cols here

        users = df.to_dict(orient='records')

        groups = {}

        ret = []

        def get_group(name):
            if name not in groups:
                groups[name] = Group(name=name)
            
            return groups[name]

        for u in users:

            c = dict(
                login_ldap = u['id'],
                email = u['email'],
                name_first = u['first'],
                name_last = u['last'],
                name_title = u['title'],
                name_nick = u['nick'],
                name_hanzi = u['hanzi'],
                tele_ext = str(u['ext']),
                tele_mobile = u['mobile'],
                # name_nick = u['nick'] if u['nick'] else None,
                )

            # c = cls(**c)

            # Membership(c, get_group(u['department']), u['order'])

            ret.append(c)

        return ret

class DiffLoader(ParamLoader):
    pass


class SQLDiffLoader(SQLLoader, DiffLoader):
    """
    Search for new items based on primary key

    Auto-add functionality if pk is missing
    (tricky it should be used from other table loaders!!!)
    TODO

    TODO Implement callback functionality to add missing entities
    """












#############################################################
# OLD LOADER based on Pandas
# It was not working mainly due to mixed types row handling difficulties

# @benchmark
# def _parse_sql(self, sql, engine):
#     """
#     TODO not in use anymore
    
#     Cannot be significantly faster as it is necessary to read data into memory
#     for conversions
#     """
#     # coerce_float=False ensures all dtypes are object_
#     with benchmark("Reading data into memory..."):
#         df = pd.read_sql(sql, engine, coerce_float=False)

#     # TODO create own describe based on pandas
#     # show more sophisticated data
#     log.debug("Describing dataframe before conversion:")
#     log.debug(df.describe(include='all'))

#     log.debug("Datatypes of dataframe:")
#     log.debug(df.dtypes)

#     # print(df)

#     col_names = [c.name for c in self.cls.__table__.columns]
#     df.columns = col_names
#     # TODO check compatibility of dimensions

#     index = []
#     for i, c in enumerate(self.cls.__table__.columns):

#         log.debug("COLUMN {} ({}) pk: {}".format(c.name, c.type, c.primary_key))
#         # df[c.name] = df[c.name].astype(str)

#         col = df[c.name] # working with one col at a time
#         col = col[pd.notnull(col)] # working with not None only

#         print(col)

#         # TODO check foreign keys for uniqueness and no None!
#         if c.foreign_keys:
#             log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))


#         if isinstance(c.type, (Date, DateTime)):
#             col = pd.to_datetime(col)
#             # TODO unrealistic date (what is realistic? (today +/- 120)
#             # series.between(left, right)

#         elif isinstance(c.type, (Integer, BigInteger, Numeric, Float)):
#             # Everything casted to Decimal to avoid precision problems, only one conversion occurs at the SQL side
#             if col.str.startswith('0').any():
#                 log.warning("Converting '0' starting strig to numeric")
#             # check dtypes and issue warnings
#             col = col.astype(dec.Decimal)
#             # pd.to_numeric(df[c.name], errors='coerce')  
#             # df.ix(pd.notnull(df[c.name]), c.name).astype(dec.Decimal) #
#             # df[c.name].replace(pd.nan, None, inplace=True)

#         elif isinstance(c.type, (String, Text, Unicode, UnicodeText)):
#             col = col.str.strip()
#             mx = col.str.len().max()  # .map(len).max() is quicker a bit
#             log.debug("COLUMN {} max length: {}".format(c.name, mx))
#             # check string and unicode columns length attribute
#             if mx > c.type.length:
#                 raise ValueError("COLUMN {} would have been truncated".format(c.name))


#         elif isinstance(c.type, (Boolean, )):
#             col = col.astype('bool')  # remove none!
#             # Allow Y-N, 0-1, True-False conversion explicitly

#         #   print(i.enums)

#         df[c.name] = col # write back parsed col
        

#     # to_numeric introduce np.nan -> monkey-patch:
#     # df = df.where(pd.notnull(df), None) # notnull selects the elements NOT replaced
#     # TODO move this to to_dict somehow as in a simple dict no np.nan should occur...

#     # show more sophisticated data
#     log.debug("Describing dataframe after conversion:")
#     log.debug(df.describe(include='all'))

#     log.debug("Datatypes of dataframe:")
#     log.debug(df.dtypes)

#     # print(df)
        
#     # convert to dictionary works only without using index
#     # df.set_index(index, inplace=True)

#     # custom df.to_dict
#     res = []
#     for df_row in df.itertuples():
#         r = {}
#         for i,c in enumerate(col_names):
#             r[c] = df_row[i]
#         res.append(r)

#     return res

#     # return df.to_dict(orient='records')  # avoid indices for this operation!



# -*- coding: utf-8 -*-
"""
This is the default implementation of the database backend of nao. It has a dependency
of sqlalchemy, but nothing sqlalchemy specific is exposed (?)

Maybe a basic sqlite implementation could also be supported here

"""

from __future__ import division, print_function, unicode_literals


from sqlalchemy.engine import create_engine
from sqlalchemy.types import Integer
from sqlalchemy.schema import Column
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base, declared_attr

#TODO: some logging facility is needed

# CONFIGURATION
# defaults
conf = {
    'host': '127.0.0.1',
    'port': 3306,
    'name': 'root',
    'pass': None,
    'db':   'test',
    }

# load YAML and update
conf.update(matrica.config['db'])
    # for k, v in params.iteritems(): params[k] = conf[k] if k in conf else v
cs = 'mysql://{name}:{pass}@{host}:{port}/{db}?charset=utf8'.format(**conf)
# log.error(cs)
engine = create_engine(cs)
# engine = create_engine('sqlite:///erih.db')
    # , echo=True
    # sqlite:///erih.db
    # server_local = r'\\VBMARI\Backup' # the runner needs to have access for that remote folder

Session = sessionmaker(bind=engine, autocommit=False, autoflush=False)
session = sess = s = Session()

# security measure to make session read-only
# session.flush = None

class DiffException(Exception): pass

class DatabaseGraph(object):
    """Creates a readable structure based on the definition

    joins realted classes etc
    """
    pass

class ModelBase(object):

    @declared_attr
    def __tablename__(cls):
        return cls.__name__

    def xml(self, columns=None, subs=None):
        cols = [c.name for c in self.__table__.c] if columns is None else columns
        base = '<{0}{1} />' if subs is None else '<{0}{1}>'
        rows = []

        rows.append(base.format(
            self.__class__.__name__,
            ' ' + ' '.join('{0}="{1}"'.format(c, getattr(self, c, None)) for c in cols),

        ))

        if subs is not None:
            rows.extend(['  '+s for s in subs])
            rows.append('</{0}>'.format(self.__class__.__name__))

        return rows

    def verbose(self)        :
        return '\n'.join(self.xml())

    def __repr__(self):
        return '<{0} #{1}>'.format(self.__class__.__name__, getattr(self, 'id', ''))


Base = declarative_base(cls=ModelBase, bind=engine)











"""
IMPLEMENTATION DRAFT

this is an ORM specific implementation it should be covered with a standard interface
the multiprocess calculation and on demand query optimization should also be standardized
to be able to use a wide variety of storage not just relational databases

that is considered as an implementation detail but it will help later on
the primary data interface exposed is that of nao Entity structure itself

there might be a native implementation (without ORM), but that would be a lot of
maybe uneccesary work
"""

from sqlalchemy import declarative_base, engine, Column, LongInteger

class PrimaryEntityStore(Base):
    """
    this table holds the basic information on entities
    
    for quick access only an alive flag is available for historized access
    maybe a birth, death double date would be more suitable
    """
    id =        Column()
    name =      Column()
    aliases =   Column()
    

class PropertyMap(Base, LifeMixin)

    id
    base_entity
    prop_entity
    value =         # comprehensive not specific type (YAML maybe)
        # distinct query on values with daterange will be needed for the entity
        # to know what are the possible values
    
    
   
"""
we need to make a collection entity that collects other entities with a weighting

class ListEntity(Entity, list): pass
class DictEntity(Entity, dict): pass
class WeightedCollection(DictEntity): pass

also introduce probability attributes for all the numeric values with
single probabilities as well as distribution-based probabilities

summing probability over time, etc.

This might also be outside the scope of nao (maybe not possible)

"""





from distutils.core import setup
setup(
  name = 'nao',
  packages = ['nao'],
  version = '0.1.7',
  description = 'Intelligent data manipulation tools',
  author = 'Szabolcs Blaga',
  author_email = 'szabolcs.blaga@gmail.com',
  url = 'https://github.com/blagasz/nao',
  download_url = 'https://github.com/blagasz/nao/tarball/0.1',
  license = 'GPL',
  install_requires=[
    'PyYAML',
    'numpy==1.11.1',
    'pandas==0.18.1',  # for datetime conversion and distinct
  ],
  keywords = ['data', 'yaml', 'multilingual', 'multivalue', 'config'],
  classifiers = [],
)


# -*- coding: utf-8 -*-
"""
Thin wrapper around PyYAML library it creates a Dumper
and Loader derived from SafeDumper and SafeLoader for 
odict and naodict data manipulation.

OdictDumer and NaoDictDumper makes block style by default.

:copyright: (c) 2014-2016 by Szabolcs Blága
:license: see LICENSE for more details   
"""

__all__ = ['naodict', 'load', 'load_all', 'dump', 'dump_all']

__version__ = '0.1-dev'

import yaml

# upstream
from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.parser import Parser
from yaml.composer import Composer
from yaml.constructor import SafeConstructor, ConstructorError
from yaml.resolver import Resolver
from yaml.representer import SafeRepresenter
from yaml.serializer import Serializer
from yaml.emitter import Emitter
# downstream

from yaml.nodes import MappingNode, ScalarNode
# from yaml.dumper import SafeDumper

try:
    from collections import OrderedDict as odict # from Python 2.7
except ImportError:
    from .od_backport import OrderedDict as odict


class naodict(odict):
    """
    Very basic attribute style access

    TODO add default `items()` like iteration
    """

    def __getattr__(self, key):
        if self._valid(key) and key in self:
            return self[key]
        else:
            return getattr(super(), key)

    def _valid(self, key):
        return not key.startswith('_')


# building ODictLoader
class ODictConstructor(SafeConstructor):

    def construct_mapping(self, node, deep=False):
        # the BaseConstructor.constructmapping is merged for simplicity
        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None,
                    "expected a mapping node, but found %s" % node.id,
                    node.start_mark)
        self.flatten_mapping(node)
        mapping = odict() # odict change
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise ConstructorError("while constructing a mapping", node.start_mark,
                        "found unacceptable key (%s)" % exc, key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping

    def construct_yaml_map(self, node):
        data = odict() # other odict change (update makes it necessary to add odict twice)
        yield data
        value = self.construct_mapping(node)
        data.update(value)

ODictConstructor.add_constructor(
        u'tag:yaml.org,2002:map',
        ODictConstructor.construct_yaml_map)


class ODictLoader(Reader, Scanner, Parser, Composer, ODictConstructor, Resolver):

    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        ODictConstructor.__init__(self)
        Resolver.__init__(self)


# building NaoDictLoader
class NaoDictConstructor(SafeConstructor):

    def construct_mapping(self, node, deep=False):
        # the BaseConstructor.constructmapping is merged for simplicity
        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None,
                    "expected a mapping node, but found %s" % node.id,
                    node.start_mark)
        self.flatten_mapping(node)
        mapping = naodict() # naodict change
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise ConstructorError("while constructing a mapping", node.start_mark,
                        "found unacceptable key (%s)" % exc, key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping

    def construct_yaml_map(self, node):
        data = naodict() # other naodict change (update makes it necessary to add odict twice)
        yield data
        value = self.construct_mapping(node)
        data.update(value)

NaoDictConstructor.add_constructor(
    u'tag:yaml.org,2002:map',
    NaoDictConstructor.construct_yaml_map)


class NaoDictLoader(Reader, Scanner, Parser, Composer, NaoDictConstructor, Resolver):

    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        NaoDictConstructor.__init__(self)
        Resolver.__init__(self)


# building ODictDumper
class ODictRepresenter(SafeRepresenter):

    def represent_mapping(self, tag, mapping, flow_style=None):
        value = []
        node = MappingNode(tag, value, flow_style=flow_style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        best_style = True
        if hasattr(mapping, 'items'):
            mapping = mapping.items()
            # mapping.sort() # the only necessary change
        for item_key, item_value in mapping:
            node_key = self.represent_data(item_key)
            node_value = self.represent_data(item_value)
            if not (isinstance(node_key, ScalarNode) and not node_key.style):
                best_style = False
            if not (isinstance(node_value, ScalarNode) and not node_value.style):
                best_style = False
            value.append((node_key, node_value))
        if flow_style is None:
            if self.default_flow_style is not None:
                node.flow_style = self.default_flow_style
            else:
                node.flow_style = best_style
        return node

ODictRepresenter.add_representer(odict,
        ODictRepresenter.represent_dict)


class ODictDumper(Emitter, Serializer, ODictRepresenter, Resolver):

    def __init__(self, stream,
            default_style=None, default_flow_style=None,
            canonical=None, indent=None, width=None,
            allow_unicode=None, line_break=None,
            encoding=None, explicit_start=None, explicit_end=None,
            version=None, tags=None):
        Emitter.__init__(self, stream, canonical=canonical,
                indent=indent, width=width,
                allow_unicode=allow_unicode, line_break=line_break)
        Serializer.__init__(self, encoding=encoding,
                explicit_start=explicit_start, explicit_end=explicit_end,
                version=version, tags=tags)
        ODictRepresenter.__init__(self, default_style=default_style,
                default_flow_style=default_flow_style)
        Resolver.__init__(self)


# building NaoDictDumper
class NaoDictRepresenter(SafeRepresenter):

    def represent_mapping(self, tag, mapping, flow_style=None):
        value = []
        node = MappingNode(tag, value, flow_style=flow_style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        best_style = True
        if hasattr(mapping, 'items'):
            mapping = mapping.items()
            # mapping.sort() # the only necessary change
        for item_key, item_value in mapping:
            node_key = self.represent_data(item_key)
            node_value = self.represent_data(item_value)
            if not (isinstance(node_key, ScalarNode) and not node_key.style):
                best_style = False
            if not (isinstance(node_value, ScalarNode) and not node_value.style):
                best_style = False
            value.append((node_key, node_value))
        if flow_style is None:
            if self.default_flow_style is not None:
                node.flow_style = self.default_flow_style
            else:
                node.flow_style = best_style
        return node

NaoDictRepresenter.add_representer(naodict,
    NaoDictRepresenter.represent_dict)


class NaoDictDumper(Emitter, Serializer, ODictRepresenter, Resolver):

    def __init__(self, stream,
            default_style=None, default_flow_style=None,
            canonical=None, indent=None, width=None,
            allow_unicode=None, line_break=None,
            encoding=None, explicit_start=None, explicit_end=None,
            version=None, tags=None):
        Emitter.__init__(self, stream, canonical=canonical,
                indent=indent, width=width,
                allow_unicode=allow_unicode, line_break=line_break)
        Serializer.__init__(self, encoding=encoding,
                explicit_start=explicit_start, explicit_end=explicit_end,
                version=version, tags=tags)
        NaoDictRepresenter.__init__(self, default_style=default_style,
                default_flow_style=default_flow_style)
        Resolver.__init__(self)


# public function overrides
_load_map = {
    'odict':    ODictLoader,
    'naodict':  NaoDictLoader,
}

def load(stream, mapping_type='odict'):
    return yaml.load(stream, _load_map[mapping_type])

def load_all(stream, mapping_type='odict'):
    return yaml.load_all(stream, _load_map[mapping_type])


_dump_map = {
    'odict':    ODictDumper,
    'naodict':  NaoDictDumper,
}

def dump(data, stream=None, mapping_type='odict', **kwds):
    kwds.setdefault('default_flow_style', False)
    return yaml.dump_all([data], stream, Dumper=_dump_map[mapping_type], **kwds)

def dump_all(documents, stream=None, mapping_type='odict', **kwds):
    kwds.setdefault('default_flow_style', False)
    return yaml.dump_all(documents, stream, Dumper=_dump_map[mapping_type], **kwds)





"""
To play with nao functionality in a finance style
interactive console

Run as ``python -m nao`` or ``pym nao`` or ``nao``
using the nao extensions for python

"""

from future.utils import python_2_unicode_compatible

from babel.numbers import format_currency

@python_2_unicode_compatible
class Currency(object):

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

M = Mil = 10**6
B = Mrd = 10**9
T = Bil = 10**12

RMB = Currency('RMB')
HUF = Currency('HUF')

def add_currencies():
    """use datebase or list or something"""

add_currencies()

base_currency = HUF

@python_2_unicode_compatible
class Amount(object):

    def __init__(self, amount, currency=base_currency):
        self.amount = amount
        self.currency = currency

    def __str__(self):
        return format_currency(self.amount, self.currency)




